package com.linuxnet.jpcsc;


/**
 * The Card-class offers the PCSC-functions related to connecting,
 * disconnecting and transfering data to cards.
 */
public final class Card{
    /**
     * Native handle.
     */
    private long card;

    /**
     * Protocol used. It is cached in a Card-instance and set
     * in Connect() and Reconnect().
     */
    private int proto;


    /**
     * Constructor.
     */
    Card(){}


    /**
     * Return protocol used.
     */
    public final int getProto(){
        return proto;
    }


    /**
     * Cleanup. 
     */
    public final void finalize(){



        // Note that the finalizer of Context could have been called
        // before and that the following call then fails.
        try{
            NativeDisconnect(PCSC.LEAVE_CARD);
        }catch(Exception e){}
    }


    /**
     * Disconnect from card.
     */
    public final void Disconnect(int param){
        int ret = NativeDisconnect(param);



        if (ret != PCSC.SUCCESS) throw new PCSCException("Disconnect()", ret);
    }


    /**
     * Reconnect to card.
     */
    public final void Reconnect(int dwSharedMode, int dwPreferredProtos, int dwInitialization){
        int ret = NativeReconnect(dwSharedMode, dwPreferredProtos, dwInitialization);




        if (ret != PCSC.SUCCESS) throw new PCSCException("Reconnect()", ret);
    }


    /**
     * Transmit given APDU. Result is returned in newly created array. 
     * In case of error, the native code raises a PCSCException.
     */
    public final byte[] Transmit(Apdu apdu){
        if (apdu == null) throw new NullPointerException();
        return Transmit(apdu.getBuffer(), 0, apdu.getLength());
    }

    /**
     * Transmit data. Result is returned in newly created array. Range within given
     * array must contain properly constructed APDU. In case of error,
     * the native code raises a PCSCException.
     */
    public final byte[] Transmit(byte[] in, int off, int len){
        if (in == null) throw new NullPointerException();
        if ((off < 0) || (len < 0) || (off + len > in.length))
            throw new ArrayIndexOutOfBoundsException();

        byte[] out = NativeTransmit(in, off, len);



        return out;
    }



    /** 
     * Transmit data to the card. Result is returned in a newly created array.
     * The given APDU header information and data are stored in a newly created
     * temporary array and passed to the native Transmit rotuines.
     * @param cla  CLA byte of APDU (byte #0)
     * @param ins  INS byte of APDU (byte #1)
     * @param p1   P1  byte of APDU (byte #2)
     * @param p2   P2  byte of APDU (byte #3)
     * @param p3   P3 or LC  byte of APDU (byte #4). This byte is omitted if parameter has value -1, otherwise it contains number of bytes in buffer in sent to card.
     * @param in   byte array containing command data of APDU. 
     * @param off  offset starting at which data sending is to begin
     * @param le   LE byte of APDU (appended to APDU). This byte is omitted if parameter has value -1.
     * @return response APDU as a byte array
     * @exception  PCSCException if communication failed or if some parameters are wrong */
    public final byte[] Transmit(int cla, int ins, int p1, int p2, int p3, byte[] in, int off, int le){
        // payload length
        int m = ((in == null) ? 0 : p3);
        // 6 = 5 byte header + 1 byte LE
        byte[] apdu = new byte[6 + m];






        apdu[0] = (byte) cla;
        apdu[1] = (byte) ins;
        apdu[2] = (byte) p1;
        apdu[3] = (byte) p2;
        if (p3 >= 0){
            apdu[4] = (byte) p3;
            if (in !=null){
                System.arraycopy(in, off, apdu, 5, p3);
                off = p3 + 5;
            }else{
                off = 5;
            }
        }else{
            off = 5;
        }
        if (le >= 0)
            apdu[off++] = (byte)le;

        return Transmit(apdu, 0, off);
    }


    /** 
     * Transmit data to the card. Result is returned in a newly created array.
     * The given APDU header information and data are stored in a newly created
     * temporary array and passed to the native Transmit rotuines. The given
     * range of the given in buffer is sent to the card.
     * @param cla  CLA byte of APDU (byte #0)
     * @param ins  INS byte of APDU (byte #1)
     * @param p1   P1  byte of APDU (byte #2)
     * @param p2   P2  byte of APDU (byte #3)
     * @param in   byte array containing command data of APDU. 
     * @param off  offset starting at which data sending is to begin
     * @param len  number of bytes to be transmitted
     * @param le   LE byte of APDU (appended to APDU). This byte is omitted if parameter has value -1.
     * @return response APDU as a byte array
     * @exception  PCSCException if communication failed or if some parameters are wrong */
    public final byte[] Transmit(int cla, int ins, int p1, int p2, byte[] in, int off, int len, int le){
        return Transmit(cla, ins, p1, p2, len, in, off, le);
    }


    /** 
     * Transmit data to the card. Result is returned in a newly created array.
     * The given APDU header information and data are stored in a newly created
     * temporary array and passed to the native Transmit rotuines. The given in
     * buffer is fully sent to the card. 
     * @param cla  CLA byte of APDU (byte #0)
     * @param ins  INS byte of APDU (byte #1)
     * @param p1   P1  byte of APDU (byte #2)
     * @param p2   P2  byte of APDU (byte #3)
     * @param in   byte array containing command data of APDU. 
     * @param le   LE byte of APDU (appended to APDU). This byte is omitted if parameter has value -1.
     * @return response APDU as a byte array
     * @exception  PCSCException if communication failed or if some parameters are wrong */
    public final byte[] Transmit(int cla, int ins, int p1, int p2, byte[] in, int le){
        return Transmit(cla, ins, p1, p2, in.length, in, 0, le);
    }



    /** 
     * Transmit data to the card. Result is returned in a newly created array.
     * The given APDU header information and data are stored in a newly created
     * temporary array and passed to the native Transmit rotuines. The given range
     * of the in buffer is sent and the expected length field is left out.
     * @param cla  CLA byte of APDU (byte #0)
     * @param ins  INS byte of APDU (byte #1)
     * @param p1   P1  byte of APDU (byte #2)
     * @param p2   P2  byte of APDU (byte #3)
     * @param in   byte array containing command data of APDU. 
     * @param off  offset starting at which data sending is to begin
     * @param len  number of bytes to be transmitted
     * @return response APDU as a byte array
     * @exception  PCSCException if communication failed or if some parameters are wrong */
    public final byte[] Transmit(int cla, int ins, int p1, int p2, byte[] in, int off, int len){
        return Transmit(cla, ins, p1, p2, len, in, off, -1);
    }



    /** 
     * Transmit data to the card. Result is returned in a newly created array.
     * The given APDU header information and data are stored in a newly created
     * temporary array and passed to the native Transmit rotuines. The in buffer
     * is sent completely, the expected length field is left out.
     * @param cla  CLA byte of APDU (byte #0)
     * @param ins  INS byte of APDU (byte #1)
     * @param p1   P1  byte of APDU (byte #2)
     * @param p2   P2  byte of APDU (byte #3)
     * @param in   byte array containing command data of APDU. 
     * @return response APDU as a byte array
     * @exception  PCSCException if communication failed or if some parameters are wrong */
    public final byte[] Transmit(int cla, int ins, int p1, int p2, byte[] in){
        return Transmit(cla, ins, p1, p2, in.length, in, 0, -1);
    }


    /**
     * Transmit data. Result is returned in the passed array at given offset. The
     * range within the given buffer in must hold a properly constructed APDU.
     * On success, the returned value will hold the number of bytes stored in the 
     * result buffer. 
     * In case of an error, a PCSCException is raised. The exception instance
     * then contains the PCSC error code. If sending the APDU was successful,
     * the number of bytes received are returned.
     */
    public final int Transmit(byte[] in, int inoff, int len, byte[] out, int outoff){
        if ((in == null) || (out == null))
            throw new NullPointerException();
        if ((inoff < 0) || (len < 0) || (inoff + len > in.length))
            throw new ArrayIndexOutOfBoundsException("Transmit(): in params");
        if ((outoff < 0) || (outoff > out.length))
            throw new ArrayIndexOutOfBoundsException("Transmit(): out params");

        int cnt = NativeTransmit(in, inoff, len, out, outoff);




        return cnt;
    }


    /**
     * Return status of connection. 
     */
    public final State Status(){
        State state = new State();



        int ret = NativeStatus(state);



        if (ret != PCSC.SUCCESS)
            throw new PCSCException("Status()", ret);

        return state;
    }


    /**
     * Signal the start of a transaction.
     */
    public final void BeginTransaction(){
        int ret = NativeBeginTransaction();



        if (ret != PCSC.SUCCESS)
            throw new PCSCException("BeginTransaction()", ret);
    }


    /**
     * Signal end of a transaction.
     */
    public final void EndTransaction(int disposition){
        int ret = NativeEndTransaction(disposition);



        if (ret != PCSC.SUCCESS)
            throw new PCSCException("EndTransaction()", ret);
    }


    /** 
     * Wrapper for SCardControl(). Note that PCSC for Linux does not
     * support the dwControlCode paramter.
     */
    public final byte[] Control(int dwControlCode, byte[] in, int off, int len){
        if ((off < 0) || (len < 0) || (off + len > in.length))
            throw new ArrayIndexOutOfBoundsException();



        return NativeControl(dwControlCode, in, off, len);
    }


    private native int NativeDisconnect(int param);

    private native byte[] NativeTransmit(byte[] in, int off, int len);

    private native int NativeTransmit(byte[] in, int inoff, int len, byte[] out, int outoff);

    private native int NativeStatus(State rs);

    private native int NativeBeginTransaction();

    private native int NativeEndTransaction(int dispo);

    private native int NativeReconnect(int dwSharedMode, int dwPreferredProtos, int dwInitialization);

    private native byte[] NativeControl(int cc, byte[] in, int off, int len);
}
