package com.linuxnet.jpcsc;


/**
 * The Context-class wraps the PCSC-functions related to 
 * connecting/disconnecting to the PCSC-service
 * and card-readers. An invocation of Connect() returns
 * a connection to a Card allowing for the transmission
 * of APDUs.
 */
public final class Context{
    /**
     * Dummy empty string array.
     */
    private static String[] emptyStringArray;

    /**
     * Static initializer.
     */
    static {
        PCSC.initialize();
    }


    /**
     * The native handle.
     */
    private long ctx;


    /**
     * Constructor. Try to use established contexts as much as possible.
     * This keeps the number of underlying connections to the PCSC-daemon
     * low.
     */
    public Context(){}


    /**
     * Cleanup operation. Destroy context we have.
     */
    public final void finalize(){



        NativeReleaseContext();
    }


    /**
     * Establish context.
     */
    public final void EstablishContext(int dwScope, String pvReserved1, String pvReserved2){
        int ret = NativeEstablishContext(dwScope, pvReserved1, pvReserved2);



        if (ret != PCSC.SUCCESS)
            throw new PCSCException("EstablishContext()", ret);
    }


    /**
     * Release context.
     */
    public final void ReleaseContext(){
        int ret = NativeReleaseContext();



        if (ret != PCSC.SUCCESS)
            throw new PCSCException("ReleaseContext()", ret);
    }


    /**
     * List all known readers in all known groups.
     */
    public final String[] ListReaders(){
        String[] sa = NativeListReaders();



        if (sa == null) return emptyStringArray;




        return sa;
    }

    /**
     * List all known readers in given group. Group may be null, all known readers by the
     * system are then returned.
     */
    public final String[] ListReaders(String group){
        String[] sa = NativeListReaders(group);



        if (sa == null) return emptyStringArray;




        return sa;
    }

    /**
     * List all known readers in given group. Group may be null, all known readers by the
     * system are then returned.
     */
    public final String[] ListReaders(String[] group){
        String[] sa = NativeListReaders(group);



        if (sa == null) return emptyStringArray;




        return sa;
    }

    /**
     * List all known reader groups by the system.
     */
    public final String[] ListReaderGroups(){
        String[] sa = NativeListReaderGroups();



        if (sa == null) return emptyStringArray;




        return sa;
    }


    /**
     * Return status change for given readers.
     */
    public final void GetStatusChange(int timeout, State[] readerStates){
        if (readerStates == null)
            throw new NullPointerException();
        int ret = NativeGetStatusChange(timeout, readerStates);







        if (ret != PCSC.SUCCESS)
            throw new PCSCException("GetStatusChange()", ret);
    }


    /**
     * Return card object in case of successful connection.
     */
    public final Card Connect(String szReader, int dwShareMode, int dwPreferredProtocols){
        if (szReader == null) throw new NullPointerException();

        Card c = new Card();
        int ret = NativeConnect(c, szReader, dwShareMode, dwPreferredProtocols);





        if (ret != PCSC.SUCCESS)
            throw new PCSCException("Connect()", ret);
        return c;
    }


    /**
     * Cancel all pending GetStatusChange() calls.
     */
    public final void Cancel(){
        int ret = NativeCancel();



        if (ret != PCSC.SUCCESS)
            throw new PCSCException("Cancel()", ret);
    }


    /**
     * Return symbolic representation of an error. Uses
     * pcsc_stringify_error() on Linux.
     */
    public native static String StringifyError(int code);

    private native int NativeEstablishContext(int dwScope, String pvReserved1, String pvReserved2);

    private native int NativeCancel();

    private native int NativeReleaseContext();

    private native String[] NativeListReaders();

    private native String[] NativeListReaders(String group);

    private native String[] NativeListReaders(String[] groups);

    private native String[] NativeListReaderGroups();

    private native int NativeGetStatusChange(int timeout, State[] readerStates);

    private native int NativeConnect(Card card, String szReader, int dwShareMode, int dwPreferredProtocols);
}
